"""
Directed Acyclic Graph of model objects.

"""
from abc import ABCMeta, abstractmethod
from collections import OrderedDict, namedtuple
from inspect import getmro

from inflection import underscore

from microcosm_postgres.cloning import clone
from microcosm_postgres.toposort import toposorted


Edge = namedtuple("Edge", ["from_id", "to_id"])


class DAG:
    """
    A graph representation using a collection of nodes and edges that supports cloning.

    """
    def __init__(self, nodes, edges=None, substitutions=None):
        """
        :param nodes: an iterable of `Model`; each must have a UUID `id`
        :param edges: an iterable of `Edge`
        :param substitutions: a dictionary of (UUID) value substitutions

        """
        self.nodes = OrderedDict(
            (node.id, node)
            for node in nodes
        )
        self.edges = edges or []
        self.substitutions = substitutions or {}

    @classmethod
    def from_nodes(cls, *nodes):
        return cls(nodes=nodes).build_edges()

    @property
    def ordered_nodes(self):
        return toposorted(self.nodes, self.edges)

    @property
    def nodes_map(self):
        """
        Build a mapping from node type to a list of nodes.

        A typed mapping helps avoid polymorphism at non-persistent layers.

        """
        dct = dict()
        for node in self.nodes.values():
            cls = next(base for base in getmro(node.__class__) if "__tablename__" in base.__dict__)
            key = getattr(cls, "__alias__", underscore(cls.__name__))
            dct.setdefault(key, []).append(node)
        return dct

    def build_edges(self):
        """
        Build edges based on node `edges` property.

        Filters out any `Edge` not defined in the DAG.

        """
        self.edges = [
            edge if isinstance(edge, Edge) else Edge(*edge)
            for node in self.nodes.values()
            for edge in getattr(node, "edges", [])
            if edge[0] in self.nodes and edge[1] in self.nodes
        ]
        return self

    def clone(self, ignore=()):
        """
        Clone this dag using a set of substitutions.

        Traverse the dag in topological order.

        """
        nodes = [
            clone(node, self.substitutions, ignore)
            for node in toposorted(self.nodes, self.edges)
        ]
        return DAG(nodes=nodes, substitutions=self.substitutions).build_edges()


class DAGCloner(metaclass=ABCMeta):
    """
    A process for building and cloning a DAG.

    The expected usage is:
      - Define a root object
      - Define a generate for child objects
      - Define extra nodes as needed
      - Clone the resulting DAG
      - Update nodes as needed

    """
    def __init__(self, graph):
        pass

    def explain(self, **kwargs):
        """
        Generate a "dry run" DAG of that state that WILL be cloned.

        """
        root = self.retrieve_root(**kwargs)
        children = self.iter_children(root, **kwargs)
        dag = DAG.from_nodes(root, *children)
        return self.add_edges(dag)

    def clone(self, substitutions, **kwargs):
        """
        Clone a DAG.

        """
        dag = self.explain(**kwargs)
        dag.substitutions.update(substitutions)
        cloned_dag = dag.clone(ignore=self.ignore)
        return self.update_nodes(self.add_edges(cloned_dag))

    @abstractmethod
    def retrieve_root(self, **kwargs):
        """
        Retrieve the root node for the DAG.

        :raises Forbidden: if the arguments to define a valid root object

        """
        pass

    @property
    def ignore(self):
        """
        Fields to ignore.

        """
        return ("created_at", "updated_at")

    def iter_children(self, root, **kwargs):
        """
        Generate child nodes for the root.

        """
        return
        yield

    def add_edges(self, dag):
        """
        Add edges using non-local node state.

        Example: import a sequential ordering based on a timestamp by adding edges between
        consecutive pairs of nodes.

        """
        return dag

    def update_nodes(self, dag):
        """
        Update nodes using non-local node state.

        Example: if one kind of node needs to update a non-UUID value based on the timestamp
        of another node.

        """
        return dag



from microcosm.api import defaults
from microcosm.scoping.factories import ScopedFactory
from sqlalchemy.orm import Session, sessionmaker


@defaults(
    engine_routing_strategy="default_engine_routing_strategy",
)
def configure_sessionmaker(graph):
    """
    Create the SQLAlchemy session class.

    """
    engine_routing_strategy = getattr(graph, graph.config.sessionmaker.engine_routing_strategy)

    if engine_routing_strategy.supports_multiple_binds:
        ScopedFactory.infect(graph, "postgres")

    class RoutingSession(Session):
        """
        Route session bind to an appropriate engine.

        See: http://docs.sqlalchemy.org/en/latest/orm/persistence_techniques.html#partitioning-strategies

        """
        def get_bind(self, mapper=None, clause=None):
            return engine_routing_strategy.get_bind(mapper, clause)

    return sessionmaker(class_=RoutingSession)



from microcosm_postgres.encryption.models import EncryptableMixin


def configure_encryptor(graph):
    """
    Create a MultiTenantEncryptor from configured keys.

    """
    encryptor = graph.multi_tenant_key_registry.make_encryptor(graph)

    # register the encryptor will each encryptable type
    for encryptable in EncryptableMixin.__subclasses__():
        encryptable.register(encryptor)

    return encryptor



"""
Compute ORM differences.

"""
from collections import namedtuple
from itertools import chain

from sqlalchemy.orm import ColumnProperty, class_mapper


Change = namedtuple("Change", ["before", "after"])


class Version(dict):
    """
    Capture of object state a specific time.

    """
    def __init__(self, instance):
        super(Version, self).__init__({
            prop.key: getattr(instance, prop.key)
            for prop in class_mapper(instance.__class__).iterate_properties
            if isinstance(prop, ColumnProperty)
        })

    def difference(self, other):
        return Delta(self, other)

    def __sub__(self, other):
        return Delta(self, other)


class Delta(dict):
    """
    Difference beween two object versions.

    """
    def __init__(self, left, right):
        super(Delta, self).__init__({
            key: Change(left.get(key), right.get(key))
            for key in set(chain(list(left.keys()), list(right.keys())))
            if left.get(key) != right.get(key)
        })



"""
Object cloning.

"""
from sqlalchemy.inspection import inspect

from microcosm_postgres.identifiers import new_object_id


def clone(instance, substitutions, ignore=()):
    """
    Clone an instance of `Model` that uses `IdentityMixin`.

    :param instance: the instance to clonse
    :param substitutions: a dictionary of substitutions
    :param ignore: a tuple of column names to ignore

    """
    substitutions[instance.id] = new_object_id()

    def substitute(value):
        try:
            hash(value)
        except Exception:
            return value
        else:
            return substitutions.get(value, value)

    return instance.__class__(**{
        key:  substitute(value)
        for key, value in iter_items(instance, ignore)
    }).create()


def iter_items(instance, ignore):
    cls = instance.__class__
    mapper = inspect(cls)
    for key, value in instance.__dict__.items():
        if key not in mapper.c:
            continue
        if key in ignore:
            continue
        yield key, value



"""
Custom types.

"""
from enum import Enum

from sqlalchemy.types import TypeDecorator, Unicode, UserDefinedType


class EnumType(TypeDecorator):
    """
    SQLAlchemy enum type that persists the enum name (not value).

    Note that this type is very similar to the `ChoiceType` from `sqlalchemy_utils`,
    with the key difference being persisting by name (and not value).

    """
    impl = Unicode(255)

    def __init__(self, enum_class):
        self.enum_class = enum_class

    @property
    def python_type(self):
        return self.impl.python_type

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        if isinstance(value, Enum):
            return str(self.enum_class(value).name)
        return str(self.enum_class[value].name)

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return self.enum_class[value]


class Serial(UserDefinedType):
    """
    A postgres Serial type.

    Intended for use with auto-incrementing immuatable columns that are NOT primary keys.

    Use in conjuction with `server_default` to ensure that SQLAlchemy fetches the generated value.

        mycolumn = Column(Serial, server_default=FetchedValue(), nullable=False)

    """
    def __init__(self, big=False):
        self.big = big

    def get_col_spec(self, **kwargs):
        """
        Column type is either SERIAL or BIGSERIAL.

        """
        return "BIGSERIAL" if self.big else "SERIAL"

    def bind_processor(self, dialect):
        """
        Always bind null to coerce auto-generation.

        """
        def process(value):
            return value
        return process

    def result_processor(self, dialect, coltype):
        """
        Always return the generated value.

        """
        def process(value):
            return value
        return process



"""
Persistence errors.

Errors define a `status_code` for each translation to HTTP (because REST)
but are not coupled with any HTTP library.

"""


class ModelIntegrityError(Exception):
    """
    An attempt to create or update a model violated a schema constraint.

    Usually the result of a programming error.

    """
    @property
    def status_code(self):
        # internal server error
        return 500


class DuplicateModelError(ModelIntegrityError):
    """
    An attempt to create or update a module violated a uniqueness constraint.

    Unlike `ModelIntegrityError`, duplicates are often expected behavior.

    """
    @property
    def status_code(self):
        # conflict
        return 409

    @property
    def include_stack_trace(self):
        return False


class MissingDependencyError(ModelIntegrityError):
    """
    An attempt to create a model didn't first create a depedency.

    """
    @property
    def status_code(self):
        # not found
        return 404

    @property
    def include_stack_trace(self):
        return False


class ReferencedModelError(ModelIntegrityError):
    """
    An attempt to delete a model didn't first delete references.

    """
    @property
    def status_code(self):
        # forbidden
        return 403

    @property
    def include_stack_trace(self):
        return False


class ModelNotFoundError(Exception):
    """
    The queried or updated model did not exist.

    """
    @property
    def status_code(self):
        # not found
        return 404

    @property
    def include_stack_trace(self):
        return False



"""
Copy a table.

"""
from sqlalchemy import Table

from microcosm_postgres.types import Serial


def copy_column(column, schema):
    """
    Safely create a copy of a column.

    """
    return column.copy(schema=schema)


def should_copy(column):
    """
    Determine if a column should be copied.

    """
    if not isinstance(column.type, Serial):
        return True

    if column.nullable:
        return True

    if not column.server_default:
        return True

    # do not create temporary serial values; they will be defaulted on upsert/insert
    return False


def copy_table(from_table, name):
    """
    Copy a table.

    Based on `Table.tometadata`, but simplified to remove constraints and indexes.

    """
    metadata = from_table.metadata

    if name in metadata.tables:
        return metadata.tables[name]

    schema = metadata.schema

    columns = [
        copy_column(column, schema)
        for column in from_table.columns
        if should_copy(column)
    ]

    return Table(
        name,
        metadata,
        schema=schema,
        comment=from_table.comment,
        *columns,
        **from_table.kwargs,
    )



"""
Alembic CLI with simplified configuration.

Alembic is a terrific tool that makes some unfortunate choices about configuration,
expecting a verbose directory structure with several layers of configuration.

This module monkey patches Alembic's CLI tool to work better within a microcosm.

To use this entry enty point instead of the Alembic CLI:

 0. Don't use `alembic init`

 1. Write models that extend `microcosm_postgres.models.Model`

 2. Add a `migrations` directory within your source tree.

    This directory does not need to be an importable Python module, but it should
    be included as part of your distribution so that migrations ship with the service.

 3. Initialize your object graph (including your models):

        from microcosm.api import create_object_graph
        from microcosm_postgres.example import Company, Employee  # noqa

        graph = create_object_graph(name="example", testing=True)

    The migrations directory is loaded by default assuming that the `name` attribute
    is a module name (though this behavior can be customized; see `microcosm.metadata:Metadata`)
    or by wiring up a stirng as the "migrations_dir" component of the graph.

 4. Write an entry point that invokes the `main` function with the object graph:

        main(graph)

"""
from contextlib import contextmanager
from os.path import isdir, join
from shutil import rmtree
from sys import argv
from tempfile import mkdtemp
from textwrap import dedent

from alembic import context
from alembic.config import CommandLine, Config
from alembic.script import ScriptDirectory
from microcosm.errors import LockedGraphError, NotBoundError

from microcosm_postgres.models import Model


def make_alembic_config(temporary_dir, migrations_dir):
    """
    Alembic uses the `alembic.ini` file to configure where it looks for everything else.

    Not only is this file an unnecessary complication around a single-valued configuration,
    the single-value it chooses to use (the alembic configuration directory), hard-coding
    the decision that there will be such a directory makes Alembic setup overly verbose.

    Instead, generate a `Config` object with the values we care about.

    :returns: a usable instance of `Alembic.config.Config`

    """
    config = Config()
    config.set_main_option("temporary_dir", temporary_dir)
    config.set_main_option("migrations_dir", migrations_dir)
    return config


def make_script_directory(cls, config):
    """
    Alembic uses a "script directory"  to encapsulate its `env.py` file, its migrations
    directory, and its `script.py.mako` revision template.

    We'd rather not have such a directory at all as the default `env.py` rarely works
    without manipulation, migrations are better saved in a location within the source tree,
    and revision templates shouldn't vary between projects.

    Instead, generate a `ScriptDirectory` object, injecting values from the config.

    """
    temporary_dir = config.get_main_option("temporary_dir")
    migrations_dir = config.get_main_option("migrations_dir")

    return cls(
        dir=temporary_dir,
        version_locations=[migrations_dir],
    )


def get_alembic_environment_options(graph):
    try:
        return graph.config.alembic.environment_options
    except (AttributeError, LockedGraphError, NotBoundError):
        return dict()


def run_online_migration(self):
    """
    Run an online migration using microcosm configuration.

    This function takes the place of the `env.py` file in the Alembic migration.

    """
    connectable = self.graph.postgres

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            # assumes that all models extend our base
            target_metadata=Model.metadata,
            **get_alembic_environment_options(self.graph),
        )

        with context.begin_transaction():
            context.run_migrations()


def make_script_py_mako():
    """
    Generate the template for new migrations.

    This function takes the place of the `script.py.mako` file in the alembic directory.

    """
    return dedent('''\
    """
    ${message}

    Revision ID: ${up_revision}
    Revises: ${down_revision | comma,n}
    Create Date: ${create_date}

    """
    from alembic import op
    import sqlalchemy as sa
    ${imports if imports else ""}

    # revision identifiers, used by Alembic.
    revision = ${repr(up_revision)}
    down_revision = ${repr(down_revision)}
    branch_labels = ${repr(branch_labels)}
    depends_on = ${repr(depends_on)}


    def upgrade():
        ${upgrades if upgrades else "pass"}


    def downgrade():
        ${downgrades if downgrades else "pass"}
    ''')


@contextmanager
def patch_script_directory(graph):
    """
    Monkey patch the `ScriptDirectory` class, working around configuration assumptions.

    Changes include:
      - Using a generated, temporary directory (with a generated, temporary `script.py.mako`)
        instead of the assumed script directory.
      - Using our `make_script_directory` function instead of the default `ScriptDirectory.from_config`.
      - Using our `run_online_migration` function instead of the default `ScriptDirectory.run_env`.
      - Injecting the current object graph.
    """
    temporary_dir = mkdtemp()
    from_config_original = getattr(ScriptDirectory, "from_config")
    run_env_original = getattr(ScriptDirectory, "run_env")

    # use a temporary directory for the revision template
    with open(join(temporary_dir, "script.py.mako"), "w") as file_:
        file_.write(make_script_py_mako())
        file_.flush()

    # monkey patch our script directory and migration logic
    setattr(ScriptDirectory, "from_config", classmethod(make_script_directory))
    setattr(ScriptDirectory, "run_env", run_online_migration)
    setattr(ScriptDirectory, "graph", graph)
    try:
        yield temporary_dir
    finally:
        # cleanup
        delattr(ScriptDirectory, "graph")
        setattr(ScriptDirectory, "run_env", run_env_original)
        setattr(ScriptDirectory, "from_config", from_config_original)
        rmtree(temporary_dir)


def get_migrations_dir(graph):
    """
    Resolve the migrations directory path.

    Either take the directory from a component of the object graph or by
    using the metaata's path resolution facilities.

    """
    try:
        migrations_dir = graph.migrations_dir
    except (LockedGraphError, NotBoundError):
        migrations_dir = graph.metadata.get_path("migrations")

    if not isdir(migrations_dir):
        raise Exception("Migrations dir must exist: {}".format(migrations_dir))
    return migrations_dir


def main(graph, *args):
    """
    Entry point for invoking Alembic's `CommandLine`.

    Alembic's CLI defines its own argument parsing and command invocation; we want
    to use these directly but define configuration our own way. This function takes
    the behavior of `CommandLine.main()` and reinterprets it with our patching.

    :param graph: an initialized object graph
    :param migration_dir: the path to the migrations directory

    """
    migrations_dir = get_migrations_dir(graph)

    cli = CommandLine()
    options = cli.parser.parse_args(args if args else argv[1:])
    if not hasattr(options, "cmd"):
        cli.parser.error("too few arguments")
    if options.cmd[0].__name__ == "init":
        cli.parser.error("Alembic 'init' command should not be used in the microcosm!")

    with patch_script_directory(graph) as temporary_dir:
        config = make_alembic_config(temporary_dir, migrations_dir)
        cli.run_cmd(config, options)



"""
Abstraction layer for persistence operations.

As much as `SQLAlchemy` provides a great deal of power, overuse of its features
creates dangerous coupling within applications. The two worst violations are:

 a. Using models directly to perform persistence operations causes persistence code
    to permeate all layers of the application, making it hard to change relationships
    and to write generic service logic in terms of a uniform persistence interface.

 b. Using explicit relationships between models makes it harder to migrate responsiblity
    for one side of the relationship to different services.

Instead, persistence operations should pass through a `Store` layer and should obey
CRUD conventions as much as possible.

"""
from contextlib import contextmanager

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import FlushError, NoResultFound

from microcosm_postgres.context import SessionContext
from microcosm_postgres.diff import Version
from microcosm_postgres.errors import (
    DuplicateModelError,
    MissingDependencyError,
    ModelIntegrityError,
    ModelNotFoundError,
    ReferencedModelError,
)
from microcosm_postgres.identifiers import new_object_id
from microcosm_postgres.metrics import postgres_metric_timing


class Store:

    def __init__(self, graph, model_class, auto_filter_fields=()):
        self.graph = graph
        self.model_class = model_class
        self.auto_filters = {
            auto_filter_field.name: auto_filter_field
            for auto_filter_field in auto_filter_fields
        }
        self.postgres_store_metrics = self.graph.postgres_store_metrics
        self.assign_model_class_store()

    def assign_model_class_store(self):
        if self.model_class:
            # Give the model class a backref to allow model-oriented CRUD
            # short cuts while still having an abstraction layer we can replace.
            self.model_class.store = self

    @property
    def model_name(self):
        return self.model_class.__name__ if self.model_class else None

    @property
    def session(self):
        return SessionContext.session

    def new_object_id(self):
        """
        Injectable id generation to facilitate mocking.

        """
        return new_object_id()

    @contextmanager
    def flushing(self):
        """
        Flush the current session, handling common errors.

        """
        try:
            yield
            self.session.flush()
        except (FlushError, IntegrityError) as error:
            error_message = str(error)
            # There ought to be a cleaner way to capture this condition
            if "duplicate" in error_message:
                raise DuplicateModelError(error)
            if "already exists" in error_message:
                raise DuplicateModelError(error)
            if "conflicts with" in error_message and "identity key" in error_message:
                raise DuplicateModelError(error)
            elif "still referenced" in error_message:
                raise ReferencedModelError(error)
            elif "is not present" in error_message:
                raise MissingDependencyError(error)
            else:
                raise ModelIntegrityError(error)

    @postgres_metric_timing(action="create")
    def create(self, instance):
        """
        Create a new model instance.

        """
        with self.flushing():
            if instance.id is None:
                instance.id = self.new_object_id()
            self.session.add(instance)
        return instance

    @postgres_metric_timing(action="retrieve")
    def retrieve(self, identifier, *criterion):
        """
        Retrieve a model by primary key and zero or more other criteria.

        :raises `NotFound` if there is no existing model

        """
        return self._retrieve(
            self.model_class.id == identifier,
            *criterion
        )

    @postgres_metric_timing(action="update")
    def update(self, identifier, new_instance):
        """
        Update an existing model with a new one.

        :raises `ModelNotFoundError` if there is no existing model

        """
        with self.flushing():
            instance = self.retrieve(identifier)
            self.merge(instance, new_instance)
            instance.updated_at = instance.new_timestamp()
        return instance

    @postgres_metric_timing(action="update_with_diff")
    def update_with_diff(self, identifier, new_instance):
        """
        Update an existing model with a new one.

        :raises `ModelNotFoundError` if there is no existing model

        """
        with self.flushing():
            instance = self.retrieve(identifier)
            before = Version(instance)
            self.merge(instance, new_instance)
            instance.updated_at = instance.new_timestamp()
            after = Version(instance)
        return instance, before - after

    def replace(self, identifier, new_instance):
        """
        Create or update a model.

        """
        try:
            # Note that `self.update()` ultimately calls merge, which will not enforce
            # a strict replacement; absent fields will default to the current values.
            return self.update(identifier, new_instance)
        except ModelNotFoundError:
            return self.create(new_instance)

    @postgres_metric_timing(action="delete")
    def delete(self, identifier):
        """
        Delete a model by primary key.

        :raises `ModelNotFoundError` if the row cannot be deleted.

        """
        return self._delete(self.model_class.id == identifier)

    @postgres_metric_timing(action="count")
    def count(self, *criterion, **kwargs):
        """
        Count the number of models matching some criterion.

        """
        query = self._query(*criterion)
        query = self._filter(query, **kwargs)
        return query.count()

    @postgres_metric_timing(action="search")
    def search(self, *criterion, **kwargs):
        """
        Return the list of models matching some criterion.

        :param offset: pagination offset, if any
        :param limit: pagination limit, if any

        """
        query = self._query(*criterion)
        query = self._order_by(query, **kwargs)
        query = self._filter(query, **kwargs)
        # NB: pagination must go last
        query = self._paginate(query, **kwargs)
        return query.all()

    @postgres_metric_timing(action="search_first")
    def search_first(self, *criterion, **kwargs):
        """
        Returns the first match based on criteria or None.

        """
        query = self._query(*criterion)
        query = self._order_by(query, **kwargs)
        query = self._filter(query, **kwargs)
        # NB: pagination must go last
        query = self._paginate(query, **kwargs)
        return query.first()

    def expunge(self, instance):
        return self.session.expunge(instance)

    def merge(self, instance, new_instance):
        self.session.merge(new_instance)

    def _order_by(self, query, **kwargs):
        """
        Add an order by clause to a (search) query.

        By default, is a noop.

        """
        return query

    def _filter(self, query, **kwargs):
        """
        Filter a query with user-supplied arguments.

        """
        query = self._auto_filter(query, **kwargs)
        return query

    def _auto_filter(self, query, **kwargs):
        for key, value in kwargs.items():
            if value is None:
                continue
            field = self.auto_filters.get(key)
            if field is None:
                continue
            query = query.filter(field == value)

        return query

    def _paginate(self, query, **kwargs):
        offset, limit = kwargs.get("offset"), kwargs.get("limit")
        if offset is not None:
            query = query.offset(offset)
        if limit is not None:
            query = query.limit(limit)

        return query

    def _retrieve(self, *criterion):
        """
        Retrieve a model by some criteria.

        :raises `ModelNotFoundError` if the row cannot be deleted.

        """
        try:
            return self._query(*criterion).one()
        except NoResultFound as error:
            raise ModelNotFoundError(
                "{} not found".format(
                    self.model_class.__name__,
                ),
                error,
            )

    def _delete(self, *criterion, synchronize_session="evaluate"):
        """
        Delete a model by some criterion.

        Avoids race-condition check-then-delete logic by checking the count of affected rows.

        NB: The `synchronize_session` keyword param is inherited from
        sqlalchemy.delete to control how sqlalchemy computes the set of rows to
        delete. This can be important for efficiency as well as flexibility. For more details see:
        https://docs.sqlalchemy.org/en/13/orm/query.html?highlight=delete#sqlalchemy.orm.query.Query.delete

        :raises `ResourceNotFound` if the row cannot be deleted.

        """
        with self.flushing():
            count = self._query(*criterion).delete(synchronize_session=synchronize_session)
        if count == 0:
            raise ModelNotFoundError
        return True

    def _query(self, *criterion):
        """
        Construct a query for the model.

        """
        return self.session.query(
            self.model_class
        ).filter(
            *criterion
        )



"""
Simple Postgres health check.

"""
from alembic.script import ScriptDirectory

from microcosm_postgres.context import SessionContext


def check_health(graph):
    """
    Basic database health check.

    """
    SessionContext.session.execute("SELECT 1;")


def check_alembic(graph):
    """
    Check connectivity to an alembic database.

    Returns the current migration.

    """
    return SessionContext.session.execute(
        "SELECT version_num FROM alembic_version LIMIT 1;",
    ).scalar()


def get_current_head_version(graph):
    """
    Returns the current head version.

    """
    script_dir = ScriptDirectory("/", version_locations=[graph.metadata.get_path("migrations")])
    return script_dir.get_current_head()



from abc import ABCMeta, abstractmethod

from microcosm.scoping.proxies import ScopedProxy


def unwrap(bind):
    return bind.__component__ if isinstance(bind, ScopedProxy) else bind


class EngineRoutingStrategy(metaclass=ABCMeta):

    @abstractmethod
    def get_bind(self, mapper, clause=None):
        pass

    @property
    @abstractmethod
    def supports_multiple_binds(self):
        pass


class DefaultEngineRoutingStrategy(EngineRoutingStrategy):
    """
    A pluggable strategy for use in our `RoutingSession`.

    """
    def __init__(self, graph):
        self.graph = graph

    def get_bind(self, mapper, clause=None):
        """
        Implement `Session.get_bind`

        """
        return unwrap(self.graph.postgres)

    @property
    def supports_multiple_binds(self):
        """
        Does this strategy support multiple concurrent engine binds?

        """
        return False


class ModelEngineRoutingStrategy(EngineRoutingStrategy):
    """
    Route based on a name embedded in a model object.

    """
    KEY = "__engine__"

    def __init__(self, graph):
        self.graph = graph

    def get_bind(self, mapper, clause=None):
        if mapper and mapper.class_:
            try:
                engine_name = getattr(mapper.class_, ModelEngineRoutingStrategy.KEY)
                with self.graph.postgres.scoped_to(engine_name):
                    return unwrap(self.graph.postgres)
            except AttributeError:
                pass

        return unwrap(self.graph.postgres)

    @property
    def supports_multiple_binds(self):
        return True



from sqlalchemy.inspection import inspect

from microcosm_postgres.encryption.models import decrypt_instance
from microcosm_postgres.store import Store


class EncryptableStore(Store):
    """
    A store for (conditionally) encryptable model.

    The store supports delete action for encryptable models by deleting
    the encrypted model.
    Note: in order to use the store, the model must define:
    -  An `encrypted_identifier` property (defaults to `encrypted_id`)
    -  An `encrypted_relationship` property (defaults to `encrypted`)

    """

    def __init__(self, graph, model_class, encrypted_store, **kwargs):
        super().__init__(graph, model_class, **kwargs)
        self.encrypted_store = encrypted_store

    def delete(self, identifier):
        instance = self.retrieve(identifier)
        result = super().delete(identifier)
        if instance.encrypted_identifier:
            self.encrypted_store.delete(instance.encrypted_identifier)
        return result

    def update(self, identifier, new_instance):
        """
        Update an encryptable field, make sure that:
        * We won't change the encryption context key
        * The new value is going to be encrypted
        * The return instance.plaintext is the updated one

        Note: Will expunge the returned instance

        """
        old_instance = self.retrieve(identifier)
        old_encrypted_identifier = old_instance.encrypted_identifier

        if (
            new_instance.encryption_context_key and
            old_instance.encryption_context_key != new_instance.encryption_context_key
        ):
            raise ValueError("Cannot change encryption context key")

        # If updating a non encrypted field - skip
        if new_instance.plaintext is None and new_instance.encrypted_relationship is None:
            result = super().update(identifier, new_instance)
            self.expunge(result)
            return result

        # Verify that the new instance is encrypted if it should be
        # If it's not - encrypt it with the old key
        # If it is - save the expected new plaintext
        if new_instance.plaintext is not None:
            expected_new_plaintext = new_instance.plaintext
            new_instance = self.reencrypt_instance(new_instance, old_instance.encryption_context_key)
        else:
            decrypt, expected_new_plaintext = decrypt_instance(new_instance)

        result = super().update(identifier, new_instance)

        # Delete the old encrypted value (instead of using sqlalchemy cascade)
        if old_encrypted_identifier != new_instance.encrypted_identifier:
            self.encrypted_store.delete(old_encrypted_identifier)

        # Update the return result, super().update() won't do it.
        self.expunge(result)
        result.plaintext = expected_new_plaintext
        return result

    def reencrypt_instance(self, instance, encryption_context_key):
        mapper = inspect(self.model_class)
        values = {
            key: value for
            key, value in instance.__dict__.items()
            if key in mapper.c
        }
        values[self.model_class.__encryption_context_key__] = encryption_context_key
        return self.model_class(**values)

    def search_encrypted_ids(self, context_id):
        query = self.session.query(
            self.model_class.id,
        ).filter(
            getattr(self.model_class, self.model_class.__encrypted_identifier__) != None,  # noqa
            getattr(self.model_class, self.model_class.__encryption_context_key__) == context_id,
        )

        return query.all()



"""
Encryption-related models.

"""
from typing import Optional, Sequence, Tuple

from sqlalchemy import Column, LargeBinary, String
from sqlalchemy.dialects.postgresql import ARRAY
from sqlalchemy.event import contains, listen, remove

from microcosm_postgres.encryption.encryptor import Encryptor


def on_init(target: "EncryptableMixin", args, kwargs):
    """
    Intercept SQLAlchemy's instance init event.

    SQLALchemy allows callback to intercept ORM instance init functions. The calling arguments
    will be an empty instance of the `target` model, plus the arguments passed to `__init__`.

    The `kwargs` dictionary is mutable (which is why it is not passed as `**kwargs`). We leverage
    this callback to conditionally remove the `__plaintext__` value and set the `ciphertext` property.

    """
    encryptor = target.__encryptor__
    assert encryptor is not None

    # encryption context may be nullable
    try:
        encryption_context_key = str(kwargs[target.__encryption_context_key__])
    except KeyError:
        return

    # do not encrypt targets that are not configured for it
    if encryption_context_key not in encryptor:
        return

    plaintext = target.plaintext_to_str(kwargs.pop(target.__plaintext__))

    # do not try to encrypt when plaintext is None
    if plaintext is None:
        return

    ciphertext, key_ids = encryptor.encrypt(encryption_context_key, plaintext)
    target.ciphertext = (ciphertext, key_ids)


def on_load(target: "EncryptableMixin", context):
    """
    Intercept SQLAlchemy's instance load event.

    """
    decrypt, plaintext = decrypt_instance(target)
    if decrypt:
        target.plaintext = plaintext  # type: ignore


def decrypt_instance(target: "EncryptableMixin") -> Tuple[bool, Optional[str]]:
    encryptor: Encryptor = target.__encryptor__  # type: ignore

    # encryption context may be nullable
    if target.encryption_context_key is None:
        return (False, None)

    encryption_context_key = str(target.encryption_context_key)

    # do not decrypt targets that are not configured for it
    if (encryption_context_key not in encryptor) or (target.ciphertext is None):
        return (False, None)

    ciphertext, key_ids = target.ciphertext
    decrypted_str = encryptor.decrypt(encryption_context_key, ciphertext)
    return (True, target.str_to_plaintext(decrypted_str))  # type: ignore


class EncryptableMixin:
    """
    A (conditionally) encryptable model.

    Using SQLAlchemy ORM events to intercept instance construction and loading to
    enforce encryption (if appropriate for the `encryption_context_key`). Should be
    combined with database constraints to enforce that the instance is *either* encrypted
    or un-encrypted, but *not* both.

    Must define:

     -  An `encryption_context_key` property (defaults to `self.key`)
     -  A settable, `plaintext` property (defaults to `self.value`)
     -  A settable, `ciphertext` property (not defaulted)

    Note: in order to use the EncryptableStore, must define:
    -  An `encrypted_identifier` property (defaults to `encrypted_id`)
    -  An `encrypted_relationship` property (defaults to `encrypted`)

    """
    __encryptor__: Optional[Encryptor] = None
    __encrypted_identifier__ = "encrypted_id"
    __encrypted_relationship__ = "encrypted"
    __encryption_context_key__ = "key"
    __plaintext__ = "value"

    @property
    def encrypted_identifier(self) -> str:
        return getattr(self, self.__encrypted_identifier__)

    @property
    def encrypted_relationship(self) -> str:
        return getattr(self, self.__encrypted_relationship__)

    @property
    def encryption_context_key(self) -> Optional[str]:
        return getattr(self, self.__encryption_context_key__)

    @property
    def plaintext(self) -> str:
        return getattr(self, self.__plaintext__)

    @plaintext.setter  # type: ignore
    def plaintext(self, value: str) -> None:
        return setattr(self, self.__plaintext__, value)

    @classmethod
    def plaintext_to_str(cls, plaintext) -> str:
        return plaintext

    @classmethod
    def str_to_plaintext(cls, text: str) -> object:
        return text

    @property
    def ciphertext(self) -> Optional[Tuple[bytes, Sequence[str]]]:
        raise NotImplementedError("Encryptable must implement `ciphertext` property")

    @ciphertext.setter
    def ciphertext(self, value: Tuple[bytes, Sequence[str]]) -> None:
        raise NotImplementedError("Encryptable must implement `ciphertext` property")

    @classmethod
    def register(cls, encryptor: Encryptor):
        """
        Register this encryptable with an encryptor.

        Instances of this encryptor will be encrypted on initialization and decrypted on load.

        """
        # save the current encryptor statically
        cls.__encryptor__ = encryptor

        # NB: we cannot use the before_insert listener in conjunction with a foreign key relationship
        # for encrypted data; SQLAlchemy will warn about using 'related attribute set' operation so
        # late in its insert/flush process.
        listeners = dict(
            init=on_init,
            load=on_load,
        )

        for name, func in listeners.items():
            # If we initialize the graph multiple times (as in many unit testing scenarios),
            # we will accumulate listener functions -- with unpredictable results. As protection,
            # we need to remove existing listeners before adding new ones; this solution only
            # works if the id (e.g. memory address) of the listener does not change, which means
            # they cannot be closures around the `encryptor` reference.
            #
            # Hence the `__encryptor__` hack above...
            if contains(cls, name, func):
                remove(cls, name, func)
            listen(cls, name, func, restore_load_context=True)


class EncryptedMixin:
    """
    A mixin that in include ciphertext and an array of key ids.

    """
    # save the encrypted data as unindexed binary
    ciphertext = Column(LargeBinary, nullable=False)
    # save the encryption key ids in an indexed column for future re-keying
    key_ids = Column(ARRAY(String), nullable=False, index=True)



"""
Topological sort.

"""
from collections import defaultdict


def toposorted(nodes, edges):
    """
    Perform a topological sort on the input resources.

    The topological sort uses Kahn's algorithm, which is a stable sort and will preserve this
    ordering; note that a DFS will produce a worst case ordering from the perspective of batching.

    """
    incoming = defaultdict(set)
    outgoing = defaultdict(set)
    for edge in edges:
        incoming[edge.to_id].add(edge.from_id)
        outgoing[edge.from_id].add(edge.to_id)

    working_set = list(nodes.values())
    results = []
    while working_set:
        remaining = []
        for node in working_set:
            if incoming[node.id]:
                # node still has incoming edges
                remaining.append(node)
                continue

            results.append(node)
            for child in outgoing[node.id]:
                incoming[child].remove(node.id)

        if len(working_set) == len(remaining):
            raise Exception("Cycle detected")
        working_set = remaining

    return results



"""
Custom key providers

"""
from os import urandom

from aws_encryption_sdk import (
    CachingCryptoMaterialsManager,
    DefaultCryptoMaterialsManager,
    LocalCryptoMaterialsCache,
)
from aws_encryption_sdk.identifiers import EncryptionKeyType, WrappingAlgorithm
from aws_encryption_sdk.internal.crypto.wrapping_keys import WrappingKey
from aws_encryption_sdk.key_providers.kms import (
    DiscoveryAwsKmsMasterKeyProvider,
    DiscoveryFilter,
    StrictAwsKmsMasterKeyProvider,
)
from aws_encryption_sdk.key_providers.raw import RawMasterKeyProvider
from microcosm.api import defaults
from microcosm.config.types import boolean
from microcosm.config.validation import typed


class StaticMasterKeyProvider(RawMasterKeyProvider):
    """
    A master key provider that uses static keys.

    Intended for unit testing.

    """
    # Declared statically to allow for multiple key_providers to be used
    # in the same unit test
    wrapping_key = urandom(32)

    @property
    def provider_id(self) -> str:
        return "static"

    def _get_raw_key(self, key_id) -> WrappingKey:
        return WrappingKey(
            wrapping_algorithm=WrappingAlgorithm.AES_256_GCM_IV12_TAG16_NO_PADDING,
            wrapping_key=self.wrapping_key,
            wrapping_key_type=EncryptionKeyType.SYMMETRIC,
        )


def configure_encrypting_key_provider(graph, key_ids):
    """
    Configure a key provider.

    During unit tests, use a static key provider (e.g. without AWS calls).

    """
    if graph.metadata.testing:
        # use static provider
        provider = StaticMasterKeyProvider()
        provider.add_master_keys_from_list(key_ids)
        return provider

    # use AWS provider
    return StrictAwsKmsMasterKeyProvider(key_ids=key_ids)


def configure_decrypting_key_provider(graph, account_ids, partition, key_ids):
    """
    Configure a key provider.

    During unit tests, use a static key provider (e.g. without AWS calls).

    """
    if graph.metadata.testing:
        # use static provider
        provider = StaticMasterKeyProvider()
        provider.add_master_keys_from_list(key_ids)
        return provider

    discovery_filter = DiscoveryFilter(
        account_ids=account_ids,
        partition=partition,
    )

    # use AWS provider
    return DiscoveryAwsKmsMasterKeyProvider(discovery_filter=discovery_filter)


@defaults(
    enable_cache=typed(boolean, default_value=False),
    cache_capacity=typed(int, default_value=100),
    cache_max_age=typed(float, default_value=3600.0),
    cache_max_messages_encrypted=typed(int, default_value=1000),
)
def configure_materials_manager(graph, key_provider):
    """
    Configure a crypto materials manager

    """
    if graph.config.materials_manager.enable_cache:
        return CachingCryptoMaterialsManager(
            cache=LocalCryptoMaterialsCache(graph.config.materials_manager.cache_capacity),
            master_key_provider=key_provider,
            max_age=graph.config.materials_manager.cache_max_age,
            max_messages_encrypted=graph.config.materials_manager.cache_max_messages_encrypted,
        )
    return DefaultCryptoMaterialsManager(master_key_provider=key_provider)



"""
Create databases.

"""
from argparse import ArgumentParser

from microcosm_postgres.operations import create_all, drop_all


def parse_args(graph):
    parser = ArgumentParser()
    parser.add_argument("--drop", "-D", action="store_true")
    # Allow services to set custom arguments for `createall`
    arguments, _ = parser.parse_known_args()
    return arguments


def main(graph):
    """
    Create and drop databases.

    """
    args = parse_args(graph)

    if args.drop:
        drop_all(graph)
    create_all(graph)



"""
Extension methods on top of a temporary table.

"""
import warnings

from sqlalchemy.dialects.postgresql import insert

from microcosm_postgres.context import SessionContext


def to_dict(item, columns):
    column_values = (
        (column, getattr(item, column.name))
        for column in columns
    )
    return dict(
        (column.name, value)
        for column, value in column_values
        # discard nulls if defaulted
        if value is not None or not column.default
    )


def insert_many(self, items):
    """
    Insert many items at once into a temporary table.

    """
    return SessionContext.session.execute(
        self.insert(values=[
            to_dict(item, self.c)
            for item in items
        ]),
    ).rowcount


def upsert_into(self, table):
    """
    Deprecated. Use `upsert_into_on_conflict_do_nothing` instead

    """
    warnings.warn(
        "`upsert_into` is deprecated. Please use `upsert_into_on_conflict_do_nothing`",
        DeprecationWarning,
    )
    return self.upsert_into_on_conflict_do_nothing(table)


def upsert_into_on_conflict_do_nothing(self, table):
    """
    Upsert from a temporarty table into another table.

    """
    return SessionContext.session.execute(
        insert(table).from_select(
            self.c,
            self,
        ).on_conflict_do_nothing(),
    ).rowcount


def upsert_into_on_conflict_do_update(self, table, **on_conflict_kwargs):
    return SessionContext.session.execute(
        insert(table).from_select(
            self.c,
            self,
        ).on_conflict_do_update(
            **on_conflict_kwargs,
        ),
    ).rowcount


def select_from(self, table):
    return SessionContext.session.query(
        table
    ).filter(
        table.id == self.c.id,
    ).all()



"""
Session context management.

"""
from contextlib import contextmanager
from functools import wraps

from microcosm_postgres.operations import new_session, recreate_all


class SessionContext:
    """
    Save current session in well-known location and provide context management.

    """
    session = None

    def __init__(self, graph, expire_on_commit=False):
        self.graph = graph
        self.expire_on_commit = expire_on_commit

    def open(self):
        SessionContext.session = new_session(self.graph, self.expire_on_commit)
        return self

    def close(self):
        if SessionContext.session:
            SessionContext.session.close()
            SessionContext.session = None

    def recreate_all(self):
        """
        Recreate all database tables, but only in a testing context.
        """
        if self.graph.metadata.testing:
            recreate_all(self.graph)

    @classmethod
    def make(cls, graph, expire_on_commit=False):
        """
        Create an opened context.

        """
        return cls(graph, expire_on_commit).open()

    # context manager

    def __enter__(self):
        return self.open()

    def __exit__(self, *args, **kwargs):
        self.close()


@contextmanager
def transaction(commit=True):
    """
    Wrap a context with a commit/rollback.

    """
    try:
        yield SessionContext.session
        if commit:
            SessionContext.session.commit()
    except Exception:
        if SessionContext.session:
            SessionContext.session.rollback()
        raise


def transactional(func):
    """
    Decorate a function call with a commit/rollback and pass the session as the first arg.

    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        with transaction():
            return func(*args, **kwargs)
    return wrapper


def maybe_transactional(func):
    """
    Variant of `transactional` that will not commit if there's an argument `commit` with a falsey value.

    Useful for dry-run style operations.

    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        commit = kwargs.get("commit", True)
        with transaction(commit=commit):
            return func(*args, **kwargs)
    return wrapper



"""
Support for building models.

Every model must inherit from `Model` and should inherit from the `EntityMixin`.

"""
from datetime import datetime
from time import time
from uuid import uuid4

from dateutil.tz import tzutc
from pytz import utc
from sqlalchemy import Column, Float, types
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy_utils import UUIDType


EPOCH = datetime(1970, 1, 1)

Model = declarative_base()


def utcnow():
    """
    Create a non-naive UTC datetime for the current time.

    Needed when *updating* UTCDateTime values because result values are currently
    converted to non-naive datetimes and SQLAlchemy cannot compare these values
    with naive datetimes generated from `datetime.utcnow()`

    """
    return datetime.now(utc)


class UTCDateTime(types.TypeDecorator):
    """
    SQLAlchemy type definition that converts stored datetime to UTC automatically.
    Source: http://stackoverflow.com/a/2528453

    """

    impl = types.DateTime

    def process_bind_param(self, value, engine):
        if value is not None:
            result = value.replace(tzinfo=None)
            return result
        else:
            return value

    def process_result_value(self, value, engine):
        if value is not None:
            result = datetime(
                value.year, value.month, value.day,
                value.hour, value.minute, value.second,
                value.microsecond, tzinfo=tzutc(),
            )
            return result
        else:
            return value


class PrimaryKeyMixin:
    """
    Define a model with a randomized UUID primary key and tracking created/updated times.

    """
    id = Column(UUIDType(), primary_key=True, default=uuid4)
    created_at = Column(UTCDateTime, default=utcnow, nullable=False)
    updated_at = Column(UTCDateTime, default=utcnow, onupdate=utcnow, nullable=False)

    def new_timestamp(self):
        return utcnow()

    @property
    def created_timestamp(self):
        return (self.created_at.replace(tzinfo=None) - EPOCH).total_seconds()

    @property
    def updated_timestamp(self):
        return (self.updated_at.replace(tzinfo=None) - EPOCH).total_seconds()


class UnixTimestampPrimaryKeyMixin:
    """
    Define a model with a randomized UUID primary key and tracking created/updated times.

    """
    id = Column(UUIDType(), primary_key=True, default=uuid4)
    created_at = Column(Float, default=time, nullable=False)
    updated_at = Column(Float, default=time, onupdate=time, nullable=False)

    def new_timestamp(self):
        return time()

    @property
    def created_timestamp(self):
        return self.created_at

    @property
    def updated_timestamp(self):
        return self.updated_at


class IdentityMixin:
    """
    Define model identity in terms of members.

    This form of equality isn't always appropriate, but it's a good place to start,
    especially for writing test assertions.

    """
    def _members(self):
        """
        Return a dict of non-private members.

        """
        return {
            key: value
            for key, value in self.__dict__.items()
            # NB: ignore internal SQLAlchemy state and nested relationships
            if not key.startswith("_") and not isinstance(value, Model)
        }

    def __eq__(self, other):
        return type(other) is type(self) and self._members() == other._members()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return id(self) if self.id is None else hash(self.id)


class SmartMixin:
    """
    Define a model with short cuts for CRUD operations against its `Store`.

    These short cuts still delegate responsibility for persistence to the store (which must be
    instantiated first).

    """
    def create(self):
        return self.__class__.store.create(self)

    def delete(self):
        return self.__class__.store.delete(self.id)

    def update(self):
        return self.__class__.store.update(self.id, self)

    def update_with_diff(self):
        return self.__class__.store.update_with_diff(self.id, self)

    def replace(self):
        return self.__class__.store.replace(self.id, self)

    @classmethod
    def search(cls, *criterion, **kwargs):
        return cls.store.search(*criterion, **kwargs)

    @classmethod
    def count(cls, *criterion):
        return cls.store.count(*criterion)

    @classmethod
    def retrieve(cls, identifier):
        return cls.store.retrieve(identifier)


class EntityMixin(PrimaryKeyMixin, IdentityMixin, SmartMixin):
    """
    Convention for persistent entities combining other mixins.

    """
    pass


class UnixTimestampEntityMixin(UnixTimestampPrimaryKeyMixin, IdentityMixin, SmartMixin):
    """
    Convention for persistent entities combining other mixins.

    """
    pass



"""
Create a SQL alchemy postgres engine.

"""
from microcosm.api import binding, defaults
from microcosm.config.types import boolean
from microcosm.config.validation import typed
from sqlalchemy import create_engine


def choose_database_name(metadata, config):
    """
    Choose the database name to use.

    As a default, databases should be named after the service that uses them. In addition,
    database names should be different between unit testing and runtime so that there is
    no chance of a unit test dropping a real database by accident.

    """
    if config.database_name is not None:
        # we allow -- but do not encourage -- database name configuration
        return config.database_name

    if metadata.testing:
        # by convention, we provision different databases for unit testing and runtime
        return f"{metadata.name}_test_db"

    return f"{metadata.name}_db"


def choose_username(metadata, config):
    """
    Choose the database username to use.

    Because databases should not be shared between services, database usernames should be
    the same as the service that uses them.

    """
    if config.username is not None:
        # we allow -- but do not encourage -- database username configuration
        return config.username

    if config.read_only:
        # by convention, we provision read-only username for every service
        return f"{metadata.name}_ro"

    return metadata.name


def choose_uri(metadata, config):
    """
    Choose the database URI to use.

    """
    database_name = choose_database_name(metadata, config)
    driver = config.driver
    host, port = config.host, config.port
    username, password = choose_username(metadata, config), config.password

    return f"{driver}://{username}:{password}@{host}:{port}/{database_name}"


def choose_connect_args(metadata, config):
    """
    Choose the SSL mode and optional root cert for the connection.

    """
    if not config.require_ssl and not config.verify_ssl:
        return dict(
            sslmode="prefer",
        )

    if config.require_ssl and not config.verify_ssl:
        return dict(
            sslmode="require",
        )

    if not config.ssl_cert_path:
        raise Exception("SSL certificate path (`ssl_cert_path`) must be configured for verification")

    return dict(
        sslmode="verify-full",
        sslrootcert=config.ssl_cert_path,
    )


def choose_args(metadata, config):
    """
    Choose database connection arguments.

    """
    return dict(
        connect_args=choose_connect_args(metadata, config),
        echo=config.echo,
        max_overflow=config.max_overflow,
        pool_size=config.pool_size,
        pool_timeout=config.pool_timeout,
    )


def make_engine(metadata, config):
    uri = choose_uri(metadata, config.postgres)
    args = choose_args(metadata, config.postgres)
    return create_engine(
        uri,
        **args,
    )


@binding("postgres")
@defaults(
    # database name will be chosen automatically if not supplied
    database_name=None,
    # database driver name; default should suffice
    driver="postgresql",
    # enable SQL echoing (verbose; only use for narrow debugging)
    echo=typed(boolean, default_value=False),
    # connection host; will usually need to be overridden (except for local development)
    host="localhost",
    # the number of extra connections over/above the pool size; 10 is the default
    max_overflow=typed(int, default_value=10),
    # password; will usually need to be overridden (except for local development)
    password="",
    # the default size of the connection pool
    pool_size=typed(int, default_value=5),
    # the timeout waiting for a connection from the pool; 30 is the default and much too large
    pool_timeout=typed(int, default_value=2),
    # the postgres connection port
    port=typed(int, default_value=5432),
    # enable read_only username convention; depends on out-of-band configuration
    read_only=typed(boolean, default_value=False),
    # use SSL to connect to postgres?
    require_ssl=typed(boolean, default_value=False),
    # username will be chosen automatically if not supplied
    username=None,
    # verify SSL connections using certificates? (requires `ssl_cert_path`)
    verify_ssl=typed(boolean, default_value=False),
    # SSL certificate path (used with `verify_ssl`)
    ssl_cert_path=None,
)
def configure_engine(graph):
    return make_engine(graph.metadata, graph.config)



"""
Implement application-layer encryption using the aws-encryption-sdk.

"""
from typing import (
    Mapping,
    Sequence,
    Tuple,
    Union,
)

from aws_encryption_sdk import CommitmentPolicy, EncryptionSDKClient
from aws_encryption_sdk.materials_managers.base import CryptoMaterialsManager


class SingleTenantEncryptor:
    """
    A single tenant encryptor.

    """
    def __init__(
        self,
        encrypting_materials_manager: CryptoMaterialsManager,
        decrypting_materials_manager: CryptoMaterialsManager,
    ):
        self.encrypting_materials_manager = encrypting_materials_manager
        self.decrypting_materials_manager = decrypting_materials_manager
        self.encryption_client = EncryptionSDKClient(
            commitment_policy=CommitmentPolicy.FORBID_ENCRYPT_ALLOW_DECRYPT,
        )

    def __contains__(self, encryption_context_key: str) -> bool:
        return True

    def encrypt(self,
                encryption_context_key: str,
                plaintext: str) -> Tuple[bytes, Sequence[str]]:
        """
        Encrypt a plaintext string value.

        The return value will include *both* the resulting binary ciphertext and the
        master key ids used for encryption. In the likely case that the encryptor was initialized
        with master key aliases, these master key ids returned will represent the unaliased key.

        """
        encryption_context = dict(
            microcosm=encryption_context_key,
        )

        ciphertext, header = self.encryption_client.encrypt(
            source=plaintext,
            materials_manager=self.encrypting_materials_manager,
            encryption_context=encryption_context,
        )

        key_ids = [
            self.unpack_key_id(encrypted_data_key.key_provider)
            for encrypted_data_key in header.encrypted_data_keys
        ]
        return ciphertext, key_ids

    def decrypt(self, encryption_context_key: str, ciphertext: bytes) -> str:
        plaintext, header = self.encryption_client.decrypt(
            source=ciphertext,
            materials_manager=self.decrypting_materials_manager,
        )
        return plaintext.decode("utf-8")

    def unpack_key_id(self, key_provider):
        key_info = key_provider.key_info
        try:
            # KMS case: the wrapped key id *is* the key id
            return key_info.decode("utf-8")
        except UnicodeDecodeError:
            # static case: the wrapped key id is the key id followed by two four byte integers (tags)
            # followed by a twelve byte initialization vectors (IV)
            #
            # see: aws_encryption_sdk.internal.formatting.serialize:serialize_wrapped_key
            return key_info[:-(4 + 4 + 12)].decode("utf-8")


class MultiTenantEncryptor:

    def __init__(self,
                 encryptors: Mapping[str, SingleTenantEncryptor],
                 default_key="default"):
        self.encryptors = encryptors
        self.default_key = default_key

    def __contains__(self, encryption_context_key: str) -> bool:
        return encryption_context_key in self.encryptors or self.default_key in self.encryptors

    def __getitem__(self, encryption_context_key: str) -> SingleTenantEncryptor:
        try:
            return self.encryptors[encryption_context_key]
        except KeyError:
            return self.encryptors[self.default_key]

    def encrypt(self, encryption_context_key: str, plaintext: str) -> Tuple[bytes, Sequence[str]]:
        encryptor = self[encryption_context_key]
        return encryptor.encrypt(encryption_context_key, plaintext)

    def decrypt(self, encryption_context_key: str, ciphertext: bytes) -> str:
        encryptor = self[encryption_context_key]
        return encryptor.decrypt(encryption_context_key, ciphertext)


Encryptor = Union[
    SingleTenantEncryptor,
    MultiTenantEncryptor,
]



#!/usr/bin/env python
from setuptools import find_packages, setup


project = "microcosm-postgres"
version = "2.2.0"

setup(
    name=project,
    version=version,
    description="Opinionated persistence with PostgreSQL",
    author="Globality Engineering",
    author_email="engineering@globality.com",
    url="https://github.com/globality-corp/microcosm-postgres",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.6",
    keywords="microcosm",
    install_requires=[
        "alembic>=1.0.0",
        "microcosm>=2.12.0",
        "microcosm-logging>=1.5.0",
        "psycopg2-binary>=2.7.5",
        "pytz>=2018.5",
        "SQLAlchemy>=1.4.10",
        "SQLAlchemy-Utils>=0.37.0",
    ],
    setup_requires=[
        "nose>=1.3.6",
    ],
    dependency_links=[
    ],
    extras_require={
        "metrics": "microcosm-metrics>=2.5.0",
        "encryption": "aws-encryption-sdk>=2.0.0",
    },
    entry_points={
        "microcosm.factories": [
            "default_engine_routing_strategy = microcosm_postgres.factories.engine_routing_strategy:DefaultEngineRoutingStrategy",  # noqa: E501
            "model_engine_routing_strategy = microcosm_postgres.factories.engine_routing_strategy:ModelEngineRoutingStrategy",  # noqa: E501
            "multi_tenant_encryptor = microcosm_postgres.encryption.factories:configure_encryptor [encryption]",
            "multi_tenant_key_registry = microcosm_postgres.encryption.registry:MultiTenantKeyRegistry [encryption]",
            "materials_manager = microcosm_postgres.encryption.providers:configure_materials_manager [encryption]",
            "postgres = microcosm_postgres.factories.engine:configure_engine",
            "postgres_store_metrics = microcosm_postgres.metrics:PostgresStoreMetrics",
            "sessionmaker = microcosm_postgres.factories.sessionmaker:configure_sessionmaker",
        ],
    },
    tests_require=[
        "coverage>=3.7.1",
        "PyHamcrest>=1.8.5",
    ],
)



"""
A registry for context keys and their master key ids.

"""
from typing import Mapping, Sequence, Union

from microcosm.api import defaults
from microcosm.config.types import comma_separated_list
from microcosm.config.validation import typed
from microcosm_logging.decorators import logger

from microcosm_postgres.encryption.encryptor import MultiTenantEncryptor, SingleTenantEncryptor
from microcosm_postgres.encryption.providers import (
    configure_decrypting_key_provider,
    configure_encrypting_key_provider,
    configure_materials_manager,
)


def parse_config(context_keys: Sequence[str],
                 key_ids: Sequence[Union[str, Sequence[str]]],
                 account_ids: Sequence[Union[str, Sequence[str]]],
                 partitions: Sequence[Union[str, Sequence[str]]]) -> Mapping[str, Mapping[str, Sequence[str]]]:
    return {
        # NB: split key id on non-comma to avoid confusion with config parsing
        context_key: {
            "key_ids": key_id.split(";") if isinstance(key_id, str) else key_id,
            "account_ids": account_id.split(";") if isinstance(account_id, str) else account_id,
            "partition": partition,
        }
        for context_key, key_id, account_id, partition in zip(context_keys, key_ids, account_ids, partitions)
    }


@defaults(
    context_keys=typed(comma_separated_list, default_value=""),
    key_ids=typed(comma_separated_list, default_value=""),
    partitions=typed(comma_separated_list, default_value=""),
    account_ids=typed(comma_separated_list, default_value=""),
)
@logger
class MultiTenantKeyRegistry:
    """
    Registry for encryption context keys and their associated master key id(s).

    """
    def __init__(self, graph):
        self.keys = parse_config(
            account_ids=graph.config.multi_tenant_key_registry.account_ids,
            context_keys=graph.config.multi_tenant_key_registry.context_keys,
            key_ids=graph.config.multi_tenant_key_registry.key_ids,
            partitions=graph.config.multi_tenant_key_registry.partitions,
        )

        for context_key, key_ids in self.keys.items():
            self.logger.info(
                "Encryption enabled for: {context_key}",
                extra=dict(
                    context_key=context_key,
                    key_ids=key_ids,
                ),
            )

    def make_encryptor(self, graph) -> MultiTenantEncryptor:
        return MultiTenantEncryptor(
            encryptors={
                context_key: SingleTenantEncryptor(
                    encrypting_materials_manager=configure_materials_manager(
                        graph,
                        key_provider=configure_encrypting_key_provider(graph, context_data["key_ids"]),
                    ),
                    decrypting_materials_manager=configure_materials_manager(
                        graph,
                        key_provider=configure_decrypting_key_provider(
                            graph,
                            context_data["account_ids"],
                            context_data["partition"],
                            context_data["key_ids"],
                        ),
                    ),
                )
                for context_key, context_data in self.keys.items()
            },
        )



"""
Transient table context manager.

"""
from contextlib import contextmanager

from microcosm_postgres.context import SessionContext
from microcosm_postgres.temporary.factories import create_transient_table


@contextmanager
def transient(from_table):
    if not SessionContext.session:
        raise Exception("transient() must be called within a SessionContext")

    transient_table = create_transient_table(from_table)

    bind = SessionContext.session.connection()
    transient_table.create(bind=bind)

    try:
        yield transient_table
    finally:
        transient_table.metadata.remove(transient_table)



"""
Common database operations.

"""
from sqlalchemy import MetaData
from sqlalchemy.exc import ProgrammingError

from microcosm_postgres.migrate import main
from microcosm_postgres.models import Model


def stamp_head(graph):
    """
    Stamp the database with the current head revision.

    """
    main(graph, "stamp", "head")


def get_current_head(graph):
    """
    Get the current database head revision, if any.

    """
    session = new_session(graph)
    try:
        result = session.execute("SELECT version_num FROM alembic_version")
    except ProgrammingError:
        return None
    else:
        return result.scalar()
    finally:
        session.close()


def create_all(graph):
    """
    Create all database tables.

    """
    head = get_current_head(graph)
    if head is None:
        Model.metadata.create_all(graph.postgres)
        stamp_head(graph)


def drop_all(graph):
    """
    Drop all database tables.

    """
    Model.metadata.drop_all(graph.postgres)
    drop_alembic_table(graph)


def drop_alembic_table(graph):
    """
    Drop the alembic version table.

    """
    try:
        graph.postgres.execute("DROP TABLE alembic_version;")
    except ProgrammingError:
        return False
    else:
        return True


# Cached database metadata instance
_metadata = None


def recreate_all(graph):
    """
    Drop and add back all database tables, or reset all data associated with a database.
    Intended mainly for testing, where a test database may either need to be re-initialized
    or cleared out between tests

    """

    global _metadata

    if _metadata is None:
        # First-run, the test database/metadata needs to be initialized
        drop_all(graph)
        create_all(graph)

        _metadata = MetaData(bind=graph.postgres)
        _metadata.reflect()
        return

    # Otherwise, truncate all existing tables
    connection = graph.postgres.connect()
    transaction = connection.begin()
    for table in reversed(_metadata.sorted_tables):
        connection.execute(table.delete())
    transaction.commit()


def new_session(graph, expire_on_commit=False):
    """
    Create a new session.

    """
    return graph.sessionmaker(expire_on_commit=expire_on_commit)



"""
Temporary table support.

"""
from types import MethodType

from microcosm_postgres.temporary.copy import copy_table
from microcosm_postgres.temporary.methods import (
    insert_many,
    select_from,
    upsert_into,
    upsert_into_on_conflict_do_nothing,
    upsert_into_on_conflict_do_update,
)


def create_temporary_table(from_table, name=None, on_commit=None):
    """
    Create a new temporary table from another table.

    """
    from_table = from_table.__table__ if hasattr(from_table, "__table__") else from_table

    name = name or f"temporary_{from_table.name}"

    # copy the origin table into the metadata with the new name.
    temporary_table = copy_table(from_table, name)

    # change create clause to: CREATE TEMPORARY TABLE
    temporary_table._prefixes = list(from_table._prefixes)
    temporary_table._prefixes.append("TEMPORARY")

    # change post create clause to: ON COMMIT DROP
    if on_commit:
        temporary_table.dialect_options["postgresql"].update(
            on_commit=on_commit,
        )

    temporary_table.insert_many = MethodType(insert_many, temporary_table)
    temporary_table.select_from = MethodType(select_from, temporary_table)

    # XXX: Deprecated. Use `upsert_into_on_conflict_do_nothing` instead
    temporary_table.upsert_into = MethodType(upsert_into, temporary_table)
    temporary_table.upsert_into_on_conflict_do_nothing = MethodType(
        upsert_into_on_conflict_do_nothing,
        temporary_table,
    )
    temporary_table.upsert_into_on_conflict_do_update = MethodType(
        upsert_into_on_conflict_do_update,
        temporary_table,
    )

    return temporary_table


def create_transient_table(from_table, name=None):
    """
    Create a transient table.

    The transient table will be dropped on commit.

    """
    return create_temporary_table(from_table, name=name, on_commit="drop")



from enum import Enum

from microcosm.api import defaults, typed
from microcosm.config.types import boolean
from microcosm.errors import NotBoundError
from microcosm_logging.timing import elapsed_time


@defaults(
    enabled=typed(boolean, default_value=True)
)
class PostgresStoreMetrics:
    """
    Send metrics regarding the postgres store

    """

    def __init__(self, graph):
        self.metrics = self.get_metrics(graph)
        self.enabled = bool(
            self.metrics
            and self.metrics.host != "localhost"
            and graph.config.postgres_store_metrics.enabled
        )

    def get_metrics(self, graph):
        """
        Fetch the metrics client from the graph.

        Metrics will be disabled if the not configured.

        """
        try:
            return graph.metrics
        except NotBoundError:
            return None

    def __call__(self, model_name, action, elapsed_time, execution_result):
        """
        Send metrics for how long it takes to execute a sql operation

        """
        if not self.enabled:
            return

        if not model_name:
            return

        key = "store"
        tags = [
            "source:microcosm-postgres",
            f"result:{execution_result}",
            f"action:{action}",
            f"model_name:{model_name}",
        ]
        self.metrics.histogram(
            key,
            elapsed_time,
            tags=tags,
        )


class SQLExecutionStatus(Enum):
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"


def postgres_metric_timing(action):
    def wrap(func):
        def func_wrapper(self, *args, **kwargs):
            extra = dict(
                model_name=self.model_name,
                action=action
            )
            try:
                with elapsed_time(extra):
                    try:
                        result = func(self, *args, **kwargs)
                        execution_status = SQLExecutionStatus.SUCCESS.name
                        return result
                    except Exception:
                        execution_status = SQLExecutionStatus.FAILURE.name
                        raise
            finally:
                self.postgres_store_metrics(
                    execution_result=execution_status,
                    **extra
                )

        return func_wrapper

    return wrap
